@@ -1,4 +1,4 @@
-%%% Copyright 2010-2011 Manolis Papadakis <manopapad@gmail.com>,
+%%% Copyright 2010-2013 Manolis Papadakis <manopapad@gmail.com>,
 %%%                     Eirini Arvaniti <eirinibob@gmail.com>
 %%%                 and Kostis Sagonas <kostis@cs.ntua.gr>
 %%%
@@ -17,7 +17,7 @@
 %%% You should have received a copy of the GNU General Public License
 %%% along with PropEr.  If not, see <http://www.gnu.org/licenses/>.
 
-%%% @copyright 2010-2011 Manolis Papadakis, Eirini Arvaniti and Kostis Sagonas
+%%% @copyright 2010-2013 Manolis Papadakis, Eirini Arvaniti and Kostis Sagonas
 %%% @version {@version}
 %%% @author Manolis Papadakis
 
@@ -245,6 +245,18 @@
 %%%   integers. This is provided as a means to speed up the testing of specs,
 %%%   where `any()' is a commonly used type (see the {@section Spec testing}
 %%%   section for details).</dd>
+%%% <dt>`{skip_mfas, [<MFA>]}'</dt>
+%%% <dd> When checking a module's specs, PropEr will not test the
+%%%   specified MFAs.  Default is [].</dd>
+%%% <dt>`{false_positive_mfas, fun((mfa(),[Arg::term()],{fail, Result::term()} | {error | exit | throw, Reason::term()}) -> boolean()) | undefined}'</dt>
+%%% <dd> When checking a module's spec(s), PropEr will treat a
+%%% counterexample as a false positive if the user supplied function
+%%% returns true.  Otherwise, PropEr will treat the counterexample as
+%%% it normally does.  The inputs to the user supplied function are
+%%% the MFA, the arguments passed to the MFA, and the result returned
+%%% from the MFA or an exception with it's reason.  If needed, the
+%%% user supplied function can call erlang:get_stacktrace/0.  Default
+%%% is undefined.</dd>
 %%% </dl>
 %%%
 %%% == Spec testing ==
@@ -348,7 +360,7 @@
 -export([pure_check/1, pure_check/2]).
 -export([forall/2, implies/2, whenfail/2, trapexit/1, timeout/2]).
 
--export_type([test/0, outer_test/0, counterexample/0, exception/0]).
+-export_type([test/0, outer_test/0, counterexample/0, exception/0, false_positive_mfas/0]).
 
 -include("proper_internal.hrl").
 
@@ -442,6 +454,7 @@
 %%-type always_clause() :: {'always', pos_integer(), delayed_test()}.
 %%-type sometimes_clause() :: {'sometimes', pos_integer(), delayed_test()}.
 
+-type false_positive_mfas() :: fun((mfa(),Args::[term()],{fail,Result::term()} | {error | exit | throw,Reason::term()}) -> boolean()) | 'undefined'.
 
 %%-----------------------------------------------------------------------------
 %% Options and Context types
@@ -462,7 +475,10 @@
 		  | {'constraint_tries',pos_integer()}
 		  | 'fails'
 		  | 'any_to_integer'
-		  | {'spec_timeout',timeout()}.
+		  | {'spec_timeout',timeout()}
+		  | {'skip_mfas', [mfa()]}
+		  | {'false_positive_mfas', false_positive_mfas()}.
+
 -type user_opts() :: [user_opt()] | user_opt().
 -record(opts, {output_fun       = fun io:format/2 :: output_fun(),
 	       long_result      = false           :: boolean(),
@@ -476,7 +492,9 @@
 	       expect_fail      = false           :: boolean(),
 	       any_type	                          :: {'type',
 						      proper_types:type()},
-	       spec_timeout     = infinity        :: timeout()}).
+	       spec_timeout     = infinity        :: timeout(),
+	       skip_mfas        = []              :: [mfa()],
+	       false_positive_mfas                :: false_positive_mfas()}).
 -type opts() :: #opts{}.
 -record(ctx, {mode     = new :: 'new' | 'try_shrunk' | 'try_cexm',
 	      bound    = []  :: imm_testcase() | counterexample(),
@@ -544,11 +562,10 @@ grow_size(#opts{max_size = MaxSize} = Opts) ->
 	true ->
 	    case get('$left') of
 		0 ->
-		    put('$size', Size + 1),
-		    case tests_at_size(Size + 1, Opts) of
-			0 -> grow_size(Opts);
-			N -> put('$left', N - 1), ok
-		    end;
+		    {ToRun, NextSize} = tests_at_next_size(Size, Opts),
+		    put('$size', NextSize),
+		    put('$left', ToRun - 1),
+		    ok;
 		Left ->
 		    put('$left', Left - 1),
 		    ok
@@ -557,24 +574,34 @@ grow_size(#opts{max_size = MaxSize} = Opts) ->
 	    ok
     end.
 
--spec tests_at_size(size(), opts()) -> non_neg_integer().
-tests_at_size(Size, #opts{numtests = NumTests, start_size = StartSize,
-			  max_size = MaxSize}) ->
+-spec tests_at_next_size(size(), opts()) -> {pos_integer(), size()}.
+tests_at_next_size(_Size, #opts{numtests = 1, start_size = StartSize}) ->
+    {1, StartSize};
+tests_at_next_size(Size, #opts{numtests = NumTests, start_size = StartSize,
+			       max_size = MaxSize})
+  when Size < MaxSize, StartSize =< MaxSize, NumTests > 1 ->
     SizesToTest = MaxSize - StartSize + 1,
     case NumTests >= SizesToTest of
 	true ->
 	    TotalOverflow = NumTests rem SizesToTest,
-	    Overflow = case MaxSize - Size < TotalOverflow of
+	    NextSize = erlang:max(StartSize, Size + 1),
+	    Overflow = case NextSize - StartSize < TotalOverflow of
 			   true  -> 1;
 			   false -> 0
 		       end,
-	    NumTests div SizesToTest + Overflow;
+	    {NumTests div SizesToTest + Overflow, NextSize};
 	false ->
-	    EverySoManySizes = SizesToTest div NumTests,
-	    case (Size - StartSize) rem EverySoManySizes of
-		0 -> 1;
-		_ -> 0
-	    end
+	    EverySoManySizes = (SizesToTest - 1) div (NumTests - 1),
+	    NextSize =
+		case Size < StartSize of
+		    true ->
+			StartSize;
+		    false ->
+			PrevMultiple =
+			    Size - (Size - StartSize) rem EverySoManySizes,
+			PrevMultiple + EverySoManySizes
+		end,
+	    {1, NextSize}
     end.
 
 %% @private
@@ -615,8 +642,8 @@ global_state_init(#opts{start_size = StartSize, constraint_tries = CTries,
         _ ->
             Seed
     end,
-    proper_arith:rand_start(S),
-    proper_typeserver:start(),
+    proper_arith:rand_restart(S),
+    proper_typeserver:restart(),
     ok.
 
 -spec global_state_reset(opts()) -> 'ok'.
@@ -639,13 +666,13 @@ global_state_erase() ->
     ok.
 
 %% @private
--spec spawn_link_migrate(fun(() -> _)) -> pid().
+-spec spawn_link_migrate(fun(() -> 'ok')) -> pid().
 spawn_link_migrate(ActualFun) ->
     PDictStuff = get(),
     Fun = fun() ->
 	      lists:foreach(fun({K,V}) -> put(K,V) end, PDictStuff),
 	      proper_arith:rand_reseed(),
-	      ActualFun()
+	      ok = ActualFun()
 	  end,
     spawn_link(Fun).
 
@@ -845,6 +872,10 @@ parse_opt(UserOpt, Opts) ->
 				    {type,proper_types:integer()}
 				};
 	{spec_timeout,N}     -> Opts#opts{spec_timeout = N};
+	{skip_mfas,L} when is_list(L)
+	                     -> Opts#opts{skip_mfas = L};
+	{false_positive_mfas,F} when is_function(F); F =:= undefined
+	                     -> Opts#opts{false_positive_mfas = F};
 	_                    -> throw({unrecognized_option,UserOpt})
     end.
 
@@ -1020,7 +1051,8 @@ retry(Test, CExm, Opts) ->
 
 -spec multi_test(mod_name(), raw_test_kind(), opts()) -> module_result().
 multi_test(Mod, RawTestKind,
-	   #opts{long_result = ReturnLong, output_fun = Print} = Opts) ->
+	   #opts{long_result = ReturnLong, output_fun = Print,
+             skip_mfas = SkipMFAs} = Opts) ->
     global_state_init(Opts),
     MaybeMFAs =
 	case RawTestKind of
@@ -1032,7 +1064,7 @@ multi_test(Mod, RawTestKind,
     {ShortResult, LongResult} =
 	case MaybeMFAs of
 	    {ok,MFAs} ->
-		RawLRes = [{MFA,mfa_test(MFA,RawTestKind,Opts)} || MFA <- MFAs],
+		RawLRes = [{MFA,mfa_test(MFA,RawTestKind,Opts)} || MFA <- MFAs--SkipMFAs],
 		LRes = [T || {_MFA,Res} = T <- RawLRes, is_list(Res)],
 		SRes = [MFA || {MFA,_Res} <- LRes],
 		save_counterexamples(LRes),
@@ -1069,8 +1101,8 @@ mfa_test({Mod,Fun,Arity} = MFA, RawTestKind, ImmOpts) ->
 -spec cook_test(raw_test(), opts()) -> test().
 cook_test({test,Test}, _Opts) ->
     Test;
-cook_test({spec,MFA}, #opts{spec_timeout = SpecTimeout}) ->
-    case proper_typeserver:create_spec_test(MFA, SpecTimeout) of
+cook_test({spec,MFA}, #opts{spec_timeout = SpecTimeout, false_positive_mfas = FalsePositiveMFAs}) ->
+    case proper_typeserver:create_spec_test(MFA, SpecTimeout, FalsePositiveMFAs) of
 	{ok,Test} ->
 	    Test;
 	{error,Reason}  ->
