@@ -182,7 +182,7 @@
 %%% Then, we parallelize a random suffix (up to 12 commands) of the initial
 %%% sequence by splitting it into 2 subsequences that will be executed
 %%% concurrently. Limitations arise from the fact that each subsequence should
-%%% be a <i>valid</i> command sequnece (i.e. all commands should satisfy
+%%% be a <i>valid</i> command sequence (i.e. all commands should satisfy
 %%% preconditions and use only symbolic variables bound to the results of
 %%% preceding calls in the same sequence). Furthermore, we apply an additional
 %%% check: we have to ensure that preconditions are satisfied in all possible
@@ -531,19 +531,15 @@ run_commands(Cmds, Env, Mod, History, State) ->
 		    case safe_apply(M2, F2, A2) of
 			{ok,Res} ->
 			    Env2 = [{V,Res}|Env],
-			    State2 = proper_symb:eval(
-				       Env2, Mod:next_state(State, Res, Call)),
 			    History2 = [{State,Res}|History],
 			    case check_postcondition(Mod, State, Call, Res) of
 				true ->
-				    run_commands(Rest, Env2, Mod, History2,
-						 State2);
+				    State2 = proper_symb:eval(Env2, Mod:next_state(State, Res, Call)),
+				    run_commands(Rest, Env2, Mod, History2, State2);
 				false ->
-				    {{lists:reverse(History2), State2,
-				      {postcondition,false}}, []};
+				    {{lists:reverse(History2), State, {postcondition,false}}, []};
 				{exception,_,_,_} = Exception ->
-				    {{lists:reverse(History2), State2,
-				      {postcondition,Exception}}, []}
+				    {{lists:reverse(History2), State, {postcondition,Exception}}, []}
 			    end;
 			{error,Exception} ->
 			    {{lists:reverse(History), State, Exception}, []}
