@@ -161,7 +161,7 @@
 -behaviour(gen_server).
 -export([demo_translate_type/2, demo_is_instance/3]).
 
--export([start/0, stop/0, create_spec_test/2, get_exp_specced/1, is_instance/3,
+-export([start/0, restart/0, stop/0, create_spec_test/3, get_exp_specced/1, is_instance/3,
 	 translate_type/1]).
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
 	 code_change/3]).
@@ -290,8 +290,9 @@
 %% @alias
 -type rich_result(T) :: {'ok',T} | {'error',term()}.
 -type rich_result2(T,S) :: {'ok',T,S} | {'error',term()}.
+-type false_positive_mfas() :: proper:false_positive_mfas().
 
--type server_call() :: {'create_spec_test',mfa(),timeout()}
+-type server_call() :: {'create_spec_test',mfa(),timeout(),false_positive_mfas()}
 		     | {'get_exp_specced',mod_name()}
 		     | {'get_type_repr',mod_name(),type_ref(),boolean()}
 		     | {'translate_type',imm_type()}.
@@ -312,6 +313,15 @@ start() ->
     put('$typeserver_pid', TypeserverPid),
     ok.
 
+%% @private
+-spec restart() -> 'ok'.
+restart() ->
+    TypeserverPid = get('$typeserver_pid'),
+    case (TypeserverPid =:= undefined orelse not is_process_alive(TypeserverPid)) of
+        true -> start();
+        false -> ok
+    end.
+
 %% @private
 -spec stop() -> 'ok'.
 stop() ->
@@ -320,10 +330,10 @@ stop() ->
     gen_server:cast(TypeserverPid, stop).
 
 %% @private
--spec create_spec_test(mfa(), timeout()) -> rich_result(proper:test()).
-create_spec_test(MFA, SpecTimeout) ->
+-spec create_spec_test(mfa(), timeout(), false_positive_mfas()) -> rich_result(proper:test()).
+create_spec_test(MFA, SpecTimeout, FalsePositiveMFAs) ->
     TypeserverPid = get('$typeserver_pid'),
-    gen_server:call(TypeserverPid, {create_spec_test,MFA,SpecTimeout}).
+    gen_server:call(TypeserverPid, {create_spec_test,MFA,SpecTimeout,FalsePositiveMFAs}).
 
 %% @private
 -spec get_exp_specced(mod_name()) -> rich_result([mfa()]).
@@ -393,8 +403,8 @@ init(_) ->
 %% @private
 -spec handle_call(server_call(), _, state()) ->
 	  {'reply',server_response(),state()}.
-handle_call({create_spec_test,MFA,SpecTimeout}, _From, State) ->
-    case create_spec_test(MFA, SpecTimeout, State) of
+handle_call({create_spec_test,MFA,SpecTimeout,FalsePositiveMFAs}, _From, State) ->
+    case create_spec_test(MFA, SpecTimeout, FalsePositiveMFAs, State) of
 	{ok,Test,NewState} ->
 	    {reply, {ok,Test}, NewState};
 	{error,_Reason} = Error ->
@@ -447,12 +457,12 @@ code_change(_OldVsn, State, _) ->
 %% Top-level interface
 %%------------------------------------------------------------------------------
 
--spec create_spec_test(mfa(), timeout(), state()) ->
+-spec create_spec_test(mfa(), timeout(), false_positive_mfas(), state()) ->
 	  rich_result2(proper:test(),state()).
-create_spec_test(MFA, SpecTimeout, State) ->
+create_spec_test(MFA, SpecTimeout, FalsePositiveMFAs, State) ->
     case get_exp_spec(MFA, State) of
 	{ok,FunRepr,NewState} ->
-	    make_spec_test(MFA, FunRepr, SpecTimeout, NewState);
+	    make_spec_test(MFA, FunRepr, SpecTimeout, FalsePositiveMFAs, NewState);
 	{error,_Reason} = Error ->
 	    Error
     end.
@@ -472,24 +482,50 @@ get_exp_spec({Mod,Fun,Arity} = MFA, State) ->
 	    Error
     end.
 
--spec make_spec_test(mfa(), fun_repr(), timeout(), state()) ->
+-spec make_spec_test(mfa(), fun_repr(), timeout(), false_positive_mfas(), state()) ->
 	  rich_result2(proper:test(),state()).
-make_spec_test({Mod,Fun,_Arity}, {Domain,Range}, SpecTimeout, State) ->
+make_spec_test({Mod,_Fun,_Arity}=MFA, {Domain,_Range}=FunRepr, SpecTimeout, FalsePositiveMFAs, State) ->
     case convert(Mod, {type,0,'$fixed_list',Domain}, State) of
 	{ok,FinType,NewState} ->
-	    %% TODO: We just catch all exceptions, plus error:badarg.
-	    Test = ?FORALL(Args, FinType, ?TIMEOUT(SpecTimeout,
-			try apply(Mod,Fun,Args) of
-			    X -> ?MODULE:is_instance(X,Mod,Range)
-			catch
-			    throw:_      -> true;
-			    error:badarg -> true
-			end)),
-	    {ok, Test, NewState};
+            Test = ?FORALL(Args, FinType, apply_spec_test(MFA, FunRepr, SpecTimeout, FalsePositiveMFAs, Args)),
+            {ok, Test, NewState};
 	{error,_Reason} = Error ->
 	    Error
     end.
 
+-spec apply_spec_test(mfa(), fun_repr(), timeout(), false_positive_mfas(), term()) -> proper:test().
+apply_spec_test({Mod,Fun,_Arity}=MFA, {_Domain,Range}, SpecTimeout, FalsePositiveMFAs, Args) ->
+    ?TIMEOUT(SpecTimeout,
+             begin
+                 %% NOTE: only call apply/3 inside try/catch (do not trust ?MODULE:is_instance/3)
+                 Result =
+                     try apply(Mod,Fun,Args) of
+                         X -> {ok, X}
+                     catch
+                         X:Y -> {X, Y}
+                     end,
+                 case Result of
+                     {ok, Z} ->
+                         case ?MODULE:is_instance(Z,Mod,Range) of
+                             true ->
+                                 true;
+                             false when is_function(FalsePositiveMFAs) ->
+                                 FalsePositiveMFAs(MFA, Args, {fail, Z});
+                             false ->
+                                 false
+                         end;
+                     Exception when is_function(FalsePositiveMFAs) ->
+                         case FalsePositiveMFAs(MFA, Args, Exception) of
+                             true ->
+                                 true;
+                             false ->
+                                 error(Exception, erlang:get_stacktrace())
+                         end;
+                     Exception ->
+                         error(Exception, erlang:get_stacktrace())
+                 end
+             end).
+
 -spec get_exp_specced(mod_name(), state()) -> rich_result2([mfa()],state()).
 get_exp_specced(Mod, State) ->
     case add_module(Mod, State) of
@@ -606,15 +642,15 @@ get_exp_info(Mod) ->
 -spec get_code_and_exports(mod_name()) ->
 	  rich_result2([abs_form()],mod_exp_funs()).
 get_code_and_exports(Mod) ->
-    case code:which(Mod) of
-	ObjFileName when is_list(ObjFileName) ->
-	    case get_chunks(ObjFileName) of
+    case code:get_object_code(Mod) of
+	{Mod, ObjBin, _ObjFileName} ->
+	    case get_chunks(ObjBin) of
 		{ok,_AbsCode,_ModExpFuns} = Result ->
 		    Result;
 		{error,Reason} ->
 		    get_code_and_exports_from_source(Mod, Reason)
 	    end;
-	_ErrAtom when is_atom(_ErrAtom) ->
+	error ->
 	    get_code_and_exports_from_source(Mod, cant_find_object_file)
     end.
 
@@ -1273,23 +1309,13 @@ is_instance(X, _Mod, {type,_,atom,[]}, _Stack) ->
 is_instance(X, _Mod, {type,_,binary,[]}, _Stack) ->
     is_binary(X);
 is_instance(X, _Mod, {type,_,binary,[BaseExpr,UnitExpr]}, _Stack) ->
+    %% <<_:X,_:_*Y>> means "bitstrings of X + k*Y bits, k >= 0"
     case eval_int(BaseExpr) of
-	{ok,0} ->
+	{ok,Base} when Base >= 0 ->
 	    case eval_int(UnitExpr) of
-		{ok,0} ->
-		    X =:= <<>>;
-		{ok,Unit} when Unit > 0 ->
-		    is_bitstring(X) andalso bit_size(X) rem Unit =:= 0;
-		_ ->
-		    abs_expr_error(invalid_unit, UnitExpr)
-	    end;
-	{ok,Len} when Len > 0 ->
-	    case eval_int(UnitExpr) of
-		{ok,0} ->
-		    %% TODO: Unspecified unit means 1-byte units?
-		    is_binary(X) andalso byte_size(X) =:= Len;
-		{ok,Unit} when Unit > 0 ->
-		    is_bitstring(X) andalso bit_size(X) =:= Len * Unit;
+		{ok,Unit} when Unit >= 0 ->
+		    is_bitstring(X) andalso bit_size(X) >= Base
+			andalso (bit_size(X) - Base) rem Unit =:= 0;
 		_ ->
 		    abs_expr_error(invalid_unit, UnitExpr)
 	    end;
@@ -1527,19 +1553,34 @@ convert(_Mod, {op,_,_Op,_Arg} = OpExpr, State, _Stack, _VarDict) ->
 convert(_Mod, {op,_,_Op,_Arg1,_Arg2} = OpExpr, State, _Stack, _VarDict) ->
     convert_integer(OpExpr, State);
 convert(_Mod, {type,_,binary,[BaseExpr,UnitExpr]}, State, _Stack, _VarDict) ->
+    %% <<_:X,_:_*Y>> means "bitstrings of X + k*Y bits, k >= 0"
     case eval_int(BaseExpr) of
 	{ok,0} ->
 	    case eval_int(UnitExpr) of
 		{ok,0} -> {ok, {simple,proper_types:exactly(<<>>)}, State};
 		{ok,1} -> {ok, {simple,proper_types:bitstring()}, State};
 		{ok,8} -> {ok, {simple,proper_types:binary()}, State};
+		{ok,N} when N > 0 ->
+		    Gen = ?LET(L, proper_types:list(proper_types:bitstring(N)),
+			       concat_bitstrings(L)),
+		    {ok, {simple,Gen}, State};
 		_      -> expr_error(invalid_unit, UnitExpr)
 	    end;
-	{ok,Len} when Len > 0 ->
+	{ok,Base} when Base > 0 ->
+	    Head = proper_types:bitstring(Base),
 	    case eval_int(UnitExpr) of
-		{ok,0} -> {ok, {simple,proper_types:binary(Len)}, State};
-		{ok,1} -> {ok, {simple,proper_types:bitstring(Len)}, State};
-		{ok,8} -> {ok, {simple,proper_types:binary(Len)}, State};
+		{ok,0} -> {ok, {simple,Head}, State};
+		{ok,1} ->
+		    Tail = proper_types:bitstring(),
+		    {ok, {simple,concat_binary_gens(Head, Tail)}, State};
+		{ok,8} ->
+		    Tail = proper_types:binary(),
+		    {ok, {simple,concat_binary_gens(Head, Tail)}, State};
+		{ok,N} when N > 0 ->
+		    Tail =
+			?LET(L, proper_types:list(proper_types:bitstring(N)),
+			     concat_bitstrings(L)),
+		    {ok, {simple,concat_binary_gens(Head, Tail)}, State};
 		_      -> expr_error(invalid_unit, UnitExpr)
 	    end;
 	_ ->
@@ -1602,6 +1643,20 @@ convert(Mod, {type,_,Name,ArgForms}, State, Stack, VarDict) ->
 convert(_Mod, TypeForm, _State, _Stack, _VarDict) ->
     {error, {unsupported_type,TypeForm}}.
 
+-spec concat_bitstrings([bitstring()]) -> bitstring().
+concat_bitstrings(BitStrings) ->
+    concat_bitstrings_tr(BitStrings, <<>>).
+
+-spec concat_bitstrings_tr([bitstring()], bitstring()) -> bitstring().
+concat_bitstrings_tr([], Acc) ->
+    Acc;
+concat_bitstrings_tr([BitString | Rest], Acc) ->
+    concat_bitstrings_tr(Rest, <<Acc/bits,BitString/bits>>).
+
+-spec concat_binary_gens(fin_type(), fin_type()) -> fin_type().
+concat_binary_gens(HeadType, TailType) ->
+    ?LET({H,T}, {HeadType,TailType}, <<H/bits,T/bits>>).
+
 -spec convert_fun(mod_name(), arity(), abs_type(), state(), stack(),
 		  var_dict()) -> rich_result2(ret_type(),state()).
 convert_fun(Mod, Arity, Range, State, Stack, VarDict) ->
